package cz.mendelu.dnaAnalyser.analyse.zdna;

import cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide;
import cz.mendelu.dnaAnalyser.sequence.stream.Window;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.regex.Pattern;

@Slf4j
@Service
public class ZdnaAnalyser {
    private static final int GCscore = 25;
    private static final int nonGCscore = 3;
    private static final int notPossibleZDNA = 0;

    /**
     * Find Z-DNAs for a given window
     *
     * @param minSequenceSize minimal length of searched sequences.
     * @param threshold minimal KV score of searched sequences.
     * @return arraylist of ZdnaAnalyseResults.
     */
    public ArrayList<ZdnaAnalyseResult> getResults(Window sequenceWindow, int minSequenceSize, float threshold) {
        ArrayList<ZdnaAnalyseResult> allResults = new ArrayList<>();

        Pattern patternGC = Pattern.compile("(?=(GC|CG))");
        Pattern patternGT = Pattern.compile("(?=(GT|TG))");

        int i = 0;
        int score = 0;
        int tmpScore;
        int len = 1;

        while (i < (sequenceWindow.getSize())) {
            tmpScore = 0;
            if (i < (sequenceWindow.getSize() - 1)) {
                Nucleotide now = sequenceWindow.get(i);
                Nucleotide next = sequenceWindow.get(i + 1);

                tmpScore = getSubScore(now, next);
            }
            if (tmpScore > 0) {
                len++;
                score += tmpScore;
            } else {
                if (len >= minSequenceSize) {
                    float kvScore = (float)score/2;
                    if (kvScore >= threshold) {
                        int start = i - len + 2;

                        StringBuilder seqB = new StringBuilder();
                        for (int k = 0; k < len; k++) {
                            seqB.append((char)sequenceWindow.get(start + k - 1).toByte());
                        }
                        String seq = seqB.toString();

                        int pcLen = len - 1;

                        // GC
                        double countGC = patternGC.matcher(seq).results().count();
                        double richnessGC = countGC / pcLen * 100;

                        // GT
                        double countGT = patternGT.matcher(seq).results().count();
                        double richnessGT = countGT / pcLen * 100;

                        ZdnaAnalyseResult zdnaResult = new ZdnaAnalyseResult(
                                null, start, len, seq, richnessGC, richnessGT, kvScore);

                        allResults.add(zdnaResult);
                    }
                }
                len = 1;
                score = 0;
            }
            i++;
        }

        return allResults;
    }

    private int getSubScore(Nucleotide first, Nucleotide second) {
        switch (first) {
            case A:
                if (second == Nucleotide.C) {
                    return nonGCscore; // AC
                }
                break;

            case T:
                if (second == Nucleotide.G) {
                    return nonGCscore; // TG
                }
                break;

            case C:
                if (second == Nucleotide.G) {
                    return GCscore; // CG
                }
                else if (second == Nucleotide.A) {
                    return nonGCscore; // CA
                }
                break;

            case G:
                if (second == Nucleotide.C) {
                    return GCscore; // GC
                }
                else if (second == Nucleotide.T) {
                    return nonGCscore; // GT
                }
                break;
        }

        return notPossibleZDNA;
    }
}
